/*
User library for a
Kernel Module to implement rdma_sendfile

Copyright (C) 2007 - Dennis Dalessandro  -  Ohio Supercomputer Center
                                         (dennis@osc.edu)

Distributed under the GNU Public License Version 2  (SEE LICENSE)
*/
#ifndef pipelinehdr
#define pipelinehdr

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <libgen.h>
#include <sys/errno.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/sendfile.h>
#include "infiniband/verbs.h"
#include "rdma/rdma_cma.h"
#include "rdma_sendfile_iface.h"

#define PIPELINE  /*Enable for pipeline rdma writes*/
#define PIPE_DYNAMIC


#if 0
#define U_RSF_DEBUG 10
#define u_rsf_debug(lvl,fmt,args...) \
    do { \
	if (lvl <= U_RSF_DEBUG) \
	    fprintf(stderr, "U_RSF DEBUG: " fmt "\n", ##args);\
    } while (0)
#else
#  define u_rsf_debug(lvl,fmt,...) do { } while (0)
#endif



typedef struct {
	int offset;
	int count;
	int current_queue;
	int free_queue;
	op_mask_t mask;
	int fd;
	uint64_t rdma_offset;
} send_args_t;

typedef user_info_t rsf_kernel_t;

typedef struct {
	struct ibv_qp 		*qp_hndl;
	struct ibv_cq 		*cq_hndl;
	struct ibv_comp_channel *cq_channel;
	struct ibv_pd		*prot_id;
	int ib_fd;
	uint32_t	rkey;
	uint64_t 	raddr;
	int		max_outstanding_ops;
	int		page_threshold;
	int		dev_fd;
	int		next_up;
	#ifdef SEND_BOGUS    /*used when we want to send gibberish ie. for doing microbenchmarks*/
		void *rdma_buf;
		uint32_t stag;
		uint64_t addr;
	#endif
	#ifdef PIPELINE
		void *rdma_buf;
		uint32_t stag;
		uint64_t addr;
	#endif
	#ifdef PIPE_DYNAMIC
		void *buf[MAX_OUTSTANDING];
		struct ibv_mr *mr[MAX_OUTSTANDING];
		int size[MAX_OUTSTANDING];
	#endif

} rsf_desc_t;


int reap_completion(rsf_desc_t *rsf_desc);
int rdma_sendfile(int fd, rsf_desc_t *rsf_desc , off_t offset, int count);

/*link QP with u_rsf data type*/
int u_rsf_init(rsf_desc_t *rsf_desc, struct ibv_qp *qp_hndl,
		struct ibv_cq *cq_hndl, struct ibv_comp_channel *cq_channel,
		uint32_t rkey, uint64_t raddr);
inline int u_rsf_fini(rsf_desc_t *rsf_desc);
inline int u_rsf_query_threshold(rsf_desc_t *rsf_desc);
inline int u_rsf_modify_threshold(rsf_desc_t *rsf_desc, int threshold);
inline int u_rsf_query_outstanding(rsf_desc_t *rsf_desc);
inline int u_rsf_modify_outstanding(rsf_desc_t *rsf_desc, int max_ops);
int do_send(rsf_desc_t *rsf_desc, send_args_t *send_args);
const char *openib_wc_status_string(int status);
#endif
