/*
Kernel Module to implement rdma_sendfile

Copyright (C) 2007 - Dennis Dalessandro  -  Ohio Supercomputer Center
                                         (dennis@osc.edu)

Distributed under the GNU Public License Version 2  (SEE LICENSE)
*/
#define RSF_KERNEL
#include "rdma_sendfile_iface.h"

#define DEVICE_NAME "rdma_sendfile"

#if 0
#define RDMA_SENDFILE_DEBUG_LEVEL 5
#define rsf_debug(lvl,fmt,args...) \
    do { \
	if (lvl <= RDMA_SENDFILE_DEBUG_LEVEL) \
	    printk(KERN_DEBUG "rdma_sendfile: " fmt "\n", ##args);\
    } while (0)
#else
#  define rsf_debug(lvl,fmt,...) do { } while (0)
#endif

typedef struct {
	user_info_t 	*user_info;
	struct ib_pd	*prot_id;
	int page_count[MAX_OUTSTANDING];
	struct page *page[MAX_OUTSTANDING][MAX_PAGE];
	struct  ib_mr *mr[MAX_OUTSTANDING];
	struct page *phys_buf_page;  /* temporary staging for phys buf list */
	#ifdef NO_REG
		struct ib_mr junk;  /*just let it get allocated*/
	#endif
}rsf_desc_t;

/*Function Prototypes*/
static int rsf_init(void);
static void rsf_exit(void);
static int rsf_open(struct inode *, struct file *);
static int rsf_release(struct inode *, struct file *);
static ssize_t rsf_read(struct file *, char *, size_t, loff_t *);
static ssize_t rsf_write(struct file *filp, const char *buff, size_t len,
			loff_t * off);
static ssize_t rsf_do_sendfile(rsf_desc_t *desc);
struct file fastcall *rsf_fget_light(unsigned int fd, int *fput_needed);
int rsf_rw_verify_area(int read_write, struct file *file, loff_t *ppos,
			size_t count);
int rsf_file_send_actor(read_descriptor_t * desc, struct page *page,
			unsigned long offset, unsigned long size);
