/*
User library for a
Kernel Module to implement rdma_sendfile

Copyright (C) 2007 - Dennis Dalessandro  -  Ohio Supercomputer Center
                                         (dennis@osc.edu)

Distributed under the GNU Public License Version 2  (SEE LICENSE)
*/

#include "u_rsf.h"
#define uint64_from_ptr(p) ((uint64_t)(uintptr_t)(p))

int reap_completion(rsf_desc_t *rsf_desc)
{
	struct ibv_wc wc;
	int sleep_time = 0;
	int polls = 100000;
	int ret;
	void *cq_context = NULL;
	struct ibv_cq *cq;

	if (rsf_desc->cq_channel == NULL) {
		/*poll off the completion queue*/
		for (;;) {
			if (polls == 0) {
				printf("Nothing showed up in CQ\n");
				return -1;
			}
			ret = ibv_poll_cq(rsf_desc->cq_hndl, 1, &wc);
			if (ret < 0) {
				printf("Error getting completion event\n");
				return -1;
			} else if (ret == 0) {
			    if (sleep_time > 0)
				usleep(sleep_time);
			} else {
			    break; /*got a completion event*/
			}
			polls --;
		}
	} else {
		/*check CQ first*/
		ret = ibv_poll_cq(rsf_desc->cq_hndl, 1, &wc);
		if (ret < 0) {
			printf("Error getting completion event\n");
			return -1;
		} else if (ret == 0) {
			/*block until get a cqe*/
BLOCK:
			u_rsf_debug(5, "Going to get event and block");
			ret = ibv_get_cq_event(rsf_desc->cq_channel, &cq,
								&cq_context);
			if(ret != 0){
				printf("Failed to get CQ event\n");
				return -1;
			}
			ibv_ack_cq_events(cq,1);
			ret = ibv_req_notify_cq(rsf_desc->cq_hndl, 0);
			if (ret) {
				printf("Unable to req notification on CQ\n");
				return -1;
			}

			ret = ibv_poll_cq(cq, 1, &wc);
			if (ret < 0) {
				printf("Did not get cq event when expectedd\n");
				return -1;
			} else if (ret == 0) {
				u_rsf_debug(5, "Got notif but nothing in CQ");
				goto BLOCK;
			}

		} else {
			u_rsf_debug(5, "nailed it on the first try");
			ret = ibv_req_notify_cq(rsf_desc->cq_hndl, 0);
			if(ret){
				printf("Unable to request notific on CQ\n");
				return -1;
			}
		}
	}

	if (wc.status != IBV_WC_SUCCESS) {
		printf("Invaild WC status received: %s\n",
					openib_wc_status_string(wc.status));
		return -1;
	}

	return wc.wr_id;

}

int rdma_sendfile(int fd, rsf_desc_t *rsf_desc , off_t offset, int count)
{

	int ret;
	int page_size = getpagesize();  /*should be 4096*/
	size_t full_round_size;
	unsigned int num_full_rounds;
	int partial_round_size=0;
	int queue_full = 0;
	int i;
	int file_pos = 0;
	int posted = 0;
	send_args_t send_args;

	send_args.fd = fd;

	full_round_size = page_size * rsf_desc->page_threshold;
	u_rsf_debug(5, "Can send at most %d bytes per time", (int)full_round_size);
	num_full_rounds = count / full_round_size;
	if(num_full_rounds * full_round_size != (unsigned)count)
		partial_round_size = count - (num_full_rounds * full_round_size);
	u_rsf_debug(5, "Need %d full rounds and %d bytes in partial round",
					num_full_rounds, partial_round_size);
	if (rsf_desc->cq_channel != NULL) {
		ret = ibv_req_notify_cq(rsf_desc->cq_hndl, 0);
		if(ret){
			printf("Unable to request notification on CQ\n");
			return -1;
		}
	}
	/*do as many full rounds as we need*/
	int send_count;
	send_count = 0;
	while(num_full_rounds > 0){
		if(queue_full == 0){
			for(i=0; i< rsf_desc->max_outstanding_ops; i++){
				if(num_full_rounds == 0)
					break;
				u_rsf_debug(5, "Reg %d no clear", ++send_count);
				send_args.offset = offset+file_pos;
				send_args.rdma_offset = rsf_desc->raddr +
								file_pos;
				send_args.count = full_round_size;
				send_args.current_queue = i;
				send_args.free_queue = NO_FREE;
				if (i == 0)
					send_args.mask = INIT | SEND_DATA;
				else
					send_args.mask = SEND_DATA;
				do_send(rsf_desc, &send_args);
				posted++;
				file_pos += full_round_size;
				num_full_rounds--;
			}
			queue_full = 1;
		} else {  /*clear one off when we reg*/
			/*first have to poll a CQ entry off*/
			ret = reap_completion(rsf_desc);
			if(ret < 0){
				printf("Did not get completion when we needed it\n");
				return -1;
			}
			if(ret != rsf_desc->next_up){
				printf("ID returned in CQ does not match next up\n");
				return -1;
			}
			send_args.offset = offset + file_pos;
			send_args.count = full_round_size;
			send_args.current_queue = rsf_desc->next_up;
			send_args.free_queue = rsf_desc->next_up;
			send_args.mask = SEND_DATA;
			send_args.rdma_offset = rsf_desc->raddr+file_pos;
			do_send(rsf_desc, &send_args);
			rsf_desc->next_up++;
			if(rsf_desc->next_up == rsf_desc->max_outstanding_ops)
				rsf_desc->next_up = 0;
			u_rsf_debug(5, "Reg and DeReg %d", ++send_count);
			file_pos += full_round_size;
			num_full_rounds--;
		}
	}
	/*need to do the parital round*/
	if(queue_full == 0 && partial_round_size > 0){
		send_args.offset = offset + file_pos;
		send_args.count = partial_round_size;
		send_args.current_queue = posted;
		send_args.free_queue = NO_FREE;
		if (posted == 0)
			send_args.mask = INIT|SEND_DATA;
		else
			send_args.mask = SEND_DATA;
		send_args.rdma_offset = rsf_desc->raddr+file_pos;
		do_send(rsf_desc, &send_args);
		posted++;
		u_rsf_debug(5, "Partial round Reg");
		file_pos += partial_round_size;
	} else if (partial_round_size > 0) {
		/*first have to poll a CQ entry off*/
		ret = reap_completion(rsf_desc);
		if(ret < 0){
			printf("Did not get completion when we needed it\n");
			return -1;
		}
		if(ret != rsf_desc->next_up){
			printf("ID returned in CQ does not match next up\n");
			return -1;
		}
		send_args.offset = offset + file_pos;
		send_args.count = partial_round_size;
		send_args.current_queue = rsf_desc->next_up;
		send_args.free_queue = rsf_desc->next_up;
		send_args.mask = SEND_DATA;
		send_args.rdma_offset = rsf_desc->raddr+file_pos;
		do_send(rsf_desc, &send_args);
		rsf_desc->next_up++;
		if(rsf_desc->next_up == rsf_desc->max_outstanding_ops)
			rsf_desc->next_up = 0;
		u_rsf_debug(5, "Partial round Reg and DeReg");
		file_pos += partial_round_size;
	}

	if(file_pos != count){
		printf("Size send %d doe snot match size requested %d", file_pos, count);
		return -1;
	}

	for(i=0; i<posted; i++){
		ret = reap_completion(rsf_desc);
		if(ret >= 0)
			u_rsf_debug(5, "Got a completion");
		else
			return -1;
	}

	return count;
}

int u_rsf_init(rsf_desc_t *rsf_desc, struct ibv_qp *qp_hndl,
		struct ibv_cq *cq_hndl, struct ibv_comp_channel *cq_channel,
		uint32_t rkey, uint64_t raddr)
{

	int ret;

	#ifdef IB
	ret = open("/dev/ib_sendfile", O_RDWR);
	#else
	ret = open("/dev/iwarp_sendfile", O_RDWR);
	#endif
	if (ret < 0) {
		perror("Can't open device\n");
		return errno;
	}
	u_rsf_debug(5, "Device is open");
	rsf_desc->dev_fd = ret;

	rsf_desc->qp_hndl = qp_hndl;
	rsf_desc->cq_hndl = cq_hndl;
	rsf_desc->rkey = rkey;
	rsf_desc->raddr = raddr;
	rsf_desc->prot_id = qp_hndl->pd;
	rsf_desc->cq_channel = cq_channel;
	//~ rsf_desc->wr_id = raddr;

	rsf_desc->page_threshold = MAX_PAGE;
	rsf_desc->max_outstanding_ops = MAX_OUTSTANDING;


	rsf_desc->next_up = 0;

	u_rsf_debug(5, "Qp=%d Rkey=%d Raddr=0x%lx", qp_hndl->qp_num, rkey,
				raddr);
	return 0;
}

inline int u_rsf_fini(rsf_desc_t *rsf_desc)
{
	int ret = close(rsf_desc->dev_fd);
	if(ret != 0){
		perror("Can't close FD");
		return errno;
	}
	u_rsf_debug(5, "Device is closed");
	return 0;

}
inline int u_rsf_query_threshold(rsf_desc_t *rsf_desc)
{
	return rsf_desc->page_threshold;
}

inline int u_rsf_modify_threshold(rsf_desc_t *rsf_desc, int threshold)
{
	if (threshold > MAX_PAGE) {
		return -1;
	}
	rsf_desc->page_threshold = threshold;
	return 0;

}

inline int u_rsf_query_outstanding(rsf_desc_t *rsf_desc)
{
	return rsf_desc->max_outstanding_ops;
}

inline int u_rsf_modify_outstanding(rsf_desc_t *rsf_desc, int max_ops)
{

	if (max_ops > MAX_OUTSTANDING) {
		return -1;
	}
	rsf_desc->max_outstanding_ops = max_ops;
	return 0;
}

int do_send(rsf_desc_t *rsf_desc, send_args_t *send_args)
{
	rsf_kernel_t kern_info;
	int dev_fd;
	uint32_t lkey;
	uint64_t addr;
	unsigned long actual_len;
	int ret;
	int kw_count;
	struct ibv_send_wr 	*bad_wr;
	struct ibv_send_wr  	rdma_wr;
	struct ibv_sge		rdma_sge;

	kw_count = sizeof(kern_info);
	kern_info.offset = send_args->offset;
	kern_info.count = send_args->count;
	kern_info.current_queue_num = send_args->current_queue;
	kern_info.free_queue_num = send_args->free_queue;
	kern_info.mask = send_args->mask;
	dev_fd = rsf_desc->dev_fd;
	kern_info.pd_hndl = rsf_desc->prot_id->handle;
	kern_info.fd = send_args->fd;
	kern_info.ib_fd = rsf_desc->ib_fd;
	kern_info.lkey = &lkey;
	kern_info.addr = &addr;
	kern_info.actual_len = &actual_len;

	/*register entire mem in one call*/
	ret = write(dev_fd, &kern_info,	kw_count);
	if (ret < 0) {
		perror("write failed");
		return -1;
	}
	if (ret != sizeof(kern_info)) {
		printf("Kernel Op Failed got %d expecting %d\n", ret, kw_count);
		return -1;
	}

	#ifdef SEND_BOGUS
		addr = rsf_desc->addr + send_args->offset;
		actual_len = send_args->count;
		lkey = rsf_desc->stag;
	#endif

	/*post the send*/
	rdma_sge.addr = addr;
	rdma_sge.length = actual_len;
	rdma_sge.lkey = lkey;
	memset(&rdma_wr, 0, sizeof(rdma_wr));
	rdma_wr.wr_id = send_args->current_queue;
	rdma_wr.sg_list = &rdma_sge;
	rdma_wr.num_sge = 1;
	rdma_wr.wr.rdma.remote_addr = send_args->rdma_offset;
	rdma_wr.wr.rdma.rkey = rsf_desc->rkey;
	rdma_wr.opcode = IBV_WR_RDMA_WRITE;
	rdma_wr.send_flags = IBV_SEND_SIGNALED;
	u_rsf_debug(5, "Posting Send");
	ret = ibv_post_send(rsf_desc->qp_hndl, &rdma_wr, &bad_wr);
	if (ret) {
		printf("Unable to post to SQ\n");
		return -1;
	}



	return 0;
}

#define CASE(e)  case e: s = #e; break
const char *openib_wc_status_string(int status)
{
    const char *s = "(UNKNOWN)";

    switch (status) {
	CASE(IBV_WC_SUCCESS);
	CASE(IBV_WC_LOC_LEN_ERR);
	CASE(IBV_WC_LOC_QP_OP_ERR);
	CASE(IBV_WC_LOC_EEC_OP_ERR);
	CASE(IBV_WC_LOC_PROT_ERR);
	CASE(IBV_WC_WR_FLUSH_ERR);
	CASE(IBV_WC_MW_BIND_ERR);
	CASE(IBV_WC_BAD_RESP_ERR);
	CASE(IBV_WC_LOC_ACCESS_ERR);
	CASE(IBV_WC_REM_INV_REQ_ERR);
	CASE(IBV_WC_REM_ACCESS_ERR);
	CASE(IBV_WC_REM_OP_ERR);
	CASE(IBV_WC_RETRY_EXC_ERR);
	CASE(IBV_WC_RNR_RETRY_EXC_ERR);
	CASE(IBV_WC_LOC_RDD_VIOL_ERR);
	CASE(IBV_WC_REM_INV_RD_REQ_ERR);
	CASE(IBV_WC_REM_ABORT_ERR);
	CASE(IBV_WC_INV_EECN_ERR);
	CASE(IBV_WC_INV_EEC_STATE_ERR);
	CASE(IBV_WC_FATAL_ERR);
	CASE(IBV_WC_GENERAL_ERR);
    }
    return s;
}

